(* Contains ppl language definition and pretty-priner *)
(* $Id: pseudo.ml,v 1.30 2000/06/25 19:39:20 alamstei Exp $ *)

open Lambda
open Library
open List

let usedlabels = ref []
let usedregs   = ref []

type register = string

type label = string

type lvalue = 
    Reg of register
  | Mem of register * int

type value =
    Lv of lvalue
  | N of int 
  | Label of label
 
type code = 
    JUMP of value
  | LOAD of lvalue * value * code
  | EXIT of string 
  | EXITOES of string * code
  | EXITOESINT of value * code
  | EXITOESBOOL of value * code
  | POP of lvalue * code
  | PUSH of value * code
  | ALLOC of lvalue * value * code
  | COMMENT of string * code
  | CMP of value * value * code
  | BE of value * code
  | BG of value * code
  | BL of value * code
  | BNE of value * code
  | BGE of value * code
  | BLE of value * code

(* ---------------------------------------------------------------- *)

(* make a fresh register *)
let new_register = 
  let counter = ref 0 in
  fun () ->
    counter := (!counter) + 1;
    "reg"^ (string_of_int !counter)

(* make a fresh label *)
let new_label =
  let counter = ref 0 in
  fun () -> 
    counter := (!counter) + 1;
    "label"^ (string_of_int !counter)

(* a pretty-printer for registers *)
let string_of_register reg =
  match reg with
    "C"  | "A" | "V" 
    | "R" | "SP" | "SS"            -> reg  (* there are special regs which are 
	        			      always present *)
  | other -> 
      usedregs := Library.add_element other (!usedregs);
      reg
	
(* a pretty-printer for labels *)
let string_of_label lab = lab

(* a pretty-printer for lvalues *)
let string_of_lvalue lval =
  match lval with
    Reg(reg) -> string_of_register reg
  | Mem(reg,place) -> ("MEM("^ (string_of_register reg) 
		       ^","^ (string_of_int place) ^")")

(* a pretty-priner for values *)
let string_of_value v = 
  match v with
    Lv(lval) -> string_of_lvalue lval
  | N(n) -> string_of_int n
  | Label(lab) -> string_of_label lab

(* a pretty-printer for codes *)
let rec string_of_code code =
  match code with
    JUMP(v) -> 
      ("     JUMP        (" ^(string_of_value v) ^")\n")
  | LOAD(lv,v,morecode) ->
      ("     LOAD        (" ^(string_of_lvalue lv) ^", "^ 
       (string_of_value v) ^")\n" ^(string_of_code morecode))
  | EXIT(str) -> 
      ("     EXIT        (\""^ str ^"\")\n")
  | EXITOES(str,morecode) -> 
      ("     EXITOES     (\""^ str ^"\")\n"^ (string_of_code morecode))
  | EXITOESINT(v,morecode) ->
      ("     EXITOESINT  ("^ (string_of_value v) ^")\n"^
       (string_of_code morecode))
  | EXITOESBOOL(v,morecode) ->
      ("     EXITOESBOOL ("^ (string_of_value v) ^")\n"^
       (string_of_code morecode))
  | POP(lv,morecode) ->
      ("     POP         ("^ (string_of_lvalue lv) ^")\n" ^
       (string_of_code morecode))
  | PUSH(v,morecode) ->
      ("     PUSH        ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
  | ALLOC(lv,v,morecode) ->
      ("     ALLOC       ("^ (string_of_lvalue lv) ^", "^
       (string_of_value v) ^")\n"^ (string_of_code morecode))
  | COMMENT(str,morecode) ->
      ("     COMMENT     ("^ str ^")\n"^ (string_of_code morecode))
  | CMP(v1,v2,morecode) ->
      ("     CMP         ("^ string_of_value v1 ^", "^ 
       (string_of_value v2) ^")\n"^ (string_of_code morecode))
  | BE(v,morecode) ->
      ("     BE          ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
  | BG(v,morecode) ->
      ("     BG          ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
  | BL(v,morecode) ->
      ("     BL          ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
  | BNE(v,morecode) ->
      ("     BNE         ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
  | BGE(v,morecode) ->
      ("     BGE         ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
  | BLE(v,morecode) ->
      ("     BLE         ("^ (string_of_value v) ^")\n" ^
       (string_of_code morecode))
 	
(* a pretty-printer for segments *)
let string_of_segment seg = 
  let (lab,segcode) = seg in
  usedlabels := Library.add_element lab (!usedlabels);
  ("LABEL("^ (string_of_label lab) ^")\n"^ (string_of_code segcode) ^"\n")

(* a pretty-printer for programs *)
let string_of_program prog = 
  let reset_globals() =
    usedlabels := [];
    usedregs := [];
  in
  reset_globals(); 
  let (main, segs) = prog in
  let mainstr = string_of_code main in
  let segsstr = Library.string_of_list string_of_segment "" segs in
  let regsstr =
    Library.string_of_list (fun x->string_of_register x^",") " " !usedregs in
  let labelstr = 
    Library.string_of_list (fun x->string_of_label x^",") " " !usedlabels in
  
  "#include \"pseudolib.h\"\n\n"^
  "REGISTERS\n     "^ regsstr ^ "\n"^
  "ENDREGISTERS\n\n"^
  "LABELS\n     "^ labelstr ^"\n     #include \"stdlabels.pc\"\n"^
  "ENDLABELS\n\n"^
  "BEGINDATA\n     #include \"stddata.pc\"\n"^
  "ENDDATA\n\n"^
  "BEGINCODE\n\n"^
    mainstr ^"\n"^
    segsstr ^"\n"^
  "#include \"stdcode.pc\"\n\n"^
  "ENDCODE\n"
