(* definition of recursive types and pretty-printer *)
(* $Id: rectypes.ml,v 1.8 2000/07/14 18:00:42 alamstei Exp $ *)

type tvar = Usertvar of string | Systemtvar of string*int
type id = int
type ptr = Var of tvar | Id of id
type node =
    Int
  | Bool
  | Arrow of ptr*ptr
  | Prod of ptr list
  | Sum of ptr list

type diagram = (id*node) list
type typeexp = ptr*diagram
type typeschema = (tvar list) * typeexp

type coloring = id list list

(* puts parenthesis around string s if bool b is true *)
let optpar b s = if b then "("^s^")" else s

(* return the free variables of a ptr *)
let free_tvars_of_ptr ptr =
  match ptr with
    Var(tvar) -> [tvar]
  | Id(id) -> []

(* return the free variables of a node *)
let free_tvars_of_node node =
  match node with
    Int -> []
  | Bool -> []
  | Arrow(ptr1,ptr2) ->
      Library.set_union (free_tvars_of_ptr ptr1) (free_tvars_of_ptr ptr2)
  | Prod(ptrlist) ->
      Library.set_flatten (List.map free_tvars_of_ptr ptrlist)
  | Sum(ptrlist) ->
      Library.set_flatten (List.map free_tvars_of_ptr ptrlist)

(* return the free variables of a diagram *)
let free_tvars_of_diagram diagram =
  let (idlist,nodelist) = List.split diagram in
  Library.set_flatten (List.map free_tvars_of_node nodelist)

(* return the free variables of a typeexp *)
let free_tvars_of_typeexp typeexp =
  let (ptr,diagram)=typeexp in
  Library.set_union (free_tvars_of_ptr ptr) (free_tvars_of_diagram diagram)

(* pretty-printer for tvars *)
let string_of_tvar tvar =
  match tvar with
    Usertvar(str) -> str
  | Systemtvar(str,int) -> str ^"!"^ (string_of_int int)

(* pretty-printer for ids *)
let string_of_id id = 
  string_of_int id

(* pretty-printer for ptrs *)
let rec string_of_ptr ptr diagram depth prec =
  match ptr with
    Var(tvar) -> string_of_tvar tvar
  | Id(id) ->
      if (depth<=0) then 
	"..."
      else
	try
	  let node = List.assoc id diagram in
	  string_of_node node diagram depth prec
	with
          (* if a labelled which is referenced is not in diag - this 
	     shouldn't happen *)
	  Not_found -> "???"

(* Pretty-printer for nodes *)
and string_of_node node diagram depth prec =
  match node with
    Arrow(b,c) -> optpar (prec>=10) 
	((string_of_ptr b diagram (depth-1) 10) ^ 
	 " -> " ^ (string_of_ptr c diagram (depth-1) 0))
  | Int -> "int"
  | Bool -> "bool"
  | Sum([]) -> "0"
  | Prod([]) -> "1"
  | Sum([t])  -> "(+ "^ string_of_ptr t diagram (depth-1) 0 ^")"
  | Prod([t]) -> "(* "^ string_of_ptr t diagram (depth-1) 0 ^")"
  | Sum(tlist) -> 
      let f = fun x -> (string_of_ptr x diagram (depth-1) 20) in
      optpar (prec>=20) (Library.string_of_list f "+" tlist)
  | Prod(tlist) ->
      let f = fun x -> (string_of_ptr x diagram (depth-1) 20) in
      optpar (prec>=20) (Library.string_of_list f "*" tlist)
	
(* pretty-printer for typeexps *)
let string_of_typeexp typeexp = 
  let (ptr,diagram) = typeexp in
  string_of_ptr ptr diagram 5 0

(* pretty-printer for typeschemas *)
let string_of_typeschema (q,a) =
  if q=[] then
    string_of_typeexp a
  else
    "forall " ^ (Library.string_of_list string_of_tvar " " q) ^
    "." ^ string_of_typeexp a

(* auxiliary function for pretty-printing "typeexp", representing any
   id's from the list "ids" as labels rather than expanding them.
   Returns pretty-printed string as well as a subset of those ids that
   were actually printed, so that the superterm knows it has to tag
   them. p is the precedence. *)
 
let rec aux_string_of_typeexp2 typeexp ids p =
  let (ptr,diag) = typeexp in
  match ptr with
    Var a -> ([],string_of_tvar a)
  | Id i ->
      if (List.mem i ids) then ([i], string_of_id i) else
      try
        let node=List.assoc i diag in
        match node with
          Arrow(p1,p2) ->
            let (ids1,str1) = aux_string_of_typeexp2 (p1,diag) (i::ids) 20 in
            let (ids2,str2) = aux_string_of_typeexp2 (p2,diag) (i::ids) 10 in
            let str = str1 ^"->"^ str2 in
            let ids' = Library.set_union ids1 ids2 in
            let str' =
              if (List.mem i ids') then
                (string_of_id i) ^":("^ str ^")"
              else
                optpar (p>=20) str
            in
            (ids',str')
	| Int -> ([], "int")
        | Bool -> ([], "bool")
 	| Sum([]) -> ([],"0")       
 	| Prod([]) -> ([],"1")
  	| Sum([t])  -> 
	    let (ids',str) = aux_string_of_typeexp2 (t,diag) (i::ids) 0 in
	    let str' =
	      if (List.mem i ids') then
		(string_of_id i) ^":("^ str ^")"
	      else
		optpar (p>=0) str 
	    in
	    (ids', "(+ "^ str' ^")")
  	| Prod([t]) -> 
	    let (ids',str) = aux_string_of_typeexp2 (t,diag) (i::ids) 0 in
	    let str' =
	      if (List.mem i ids') then
		(string_of_id i) ^":("^ str ^")"
	      else
		optpar (p>=0) str 
	    in
	    (ids', "(* "^ str' ^")")
  	| Sum(tlist) -> 
	    let f = fun x -> aux_string_of_typeexp2 (x,diag) (i::ids) 20 in
	    let (idlist,strlist) = List.split (List.map f tlist) in
	    let ids' = Library.set_flatten idlist in
	    let str = Library.string_of_list (fun x -> x) "+" strlist in
	    let str' =
	      if (List.mem i ids') then
		(string_of_id i) ^":("^ str ^")"
	      else
		optpar (p>=0) str 
	    in
	    (ids', str')
  	| Prod(tlist) ->
 	    let f = fun x -> aux_string_of_typeexp2 (x,diag) (i::ids) 20 in
	    let (idlist,strlist) = List.split (List.map f tlist) in
	    let ids' = Library.set_flatten idlist in
	    let str = Library.string_of_list (fun x -> x) "*" strlist in
	    let str' =
	      if (List.mem i ids') then
		(string_of_id i) ^":("^ str ^")"
	      else
		optpar (p>=0) str 
	    in
	    (ids', str')
      with Not_found -> (* if an id is not in diag - this shouldn't happen *)
        ([],"???")
 
(* pretty-printer for typeexps.  Any cyclic reference to nodes results in the 
   printing of those nodes by their identifiers. *)
let string_of_typeexp2 typeexp =
  let (usedids,str) = aux_string_of_typeexp2 typeexp [] 0 in
  str

(* return true iff two ids are of the same color *)
let equivalent coloring id1 id2 =
  try
    let color = List.find (List.mem id1) coloring in
    List.mem id2 color
  with
    Not_found -> false (* raised by List.find *)

(* return true iff two pointers are equivalent with respect to a coloring *)
let equivalent_ptr coloring ptr1 ptr2 =
  match (ptr1, ptr2) with
    (Var(tvar1),Var(tvar2)) -> 
      tvar1=tvar2 
  | (Id(id1),Id(id2)) -> 
      equivalent coloring id1 id2
  | _ -> false

(* return true iff two nodes are equivalent with respect to a coloring *)
let equivalent_node coloring node1 node2 =
  match (node1,node2) with
    (Int,Int) -> true
  | (Bool,Bool) -> true
  | (Arrow(ptr1,ptr2),Arrow(ptr3,ptr4)) ->
      (equivalent_ptr coloring ptr1 ptr3) && 
      (equivalent_ptr coloring ptr2 ptr4)
  | (Prod(ptrlist1),Prod(ptrlist2)) ->
      begin
	try
	  let f = fun (a,b) -> equivalent_ptr coloring a b in
          List.for_all f (List.combine ptrlist1 ptrlist2)
      	with
          Invalid_argument _ -> false (* raised by List.combine *)  
      end      
  | (Sum(ptrlist1),Sum(ptrlist2)) ->
      begin      
	try
	  let f = fun (a,b) -> equivalent_ptr coloring a b in
          List.for_all f (List.combine ptrlist1 ptrlist2)
      	with
          Invalid_argument _ -> false (* raised by List.combine *)  
      end      
  | _ -> false
	
(* takes a colored diagram and a non-empty list of id's.  It looks at the
   right-hand-side of each id in diagram, and separates the ids into
   groups according to whether the corresponding nodes are equivalent for
   the given coloring (in the sense of equivalent_node above). It returns
   a new list of id lists. *)
let rec split_colors diag col ids =
  match ids with
    [] -> []  (* base case *)
  | i0::t ->
      let node0 = List.assoc i0 diag in
      let split_fun i =
	try
	  equivalent_node col (List.assoc i diag) node0
	with
	  Not_found -> false
      in
      (* find all ids whose nodes are equivalent to node0, and those
         that are different *)
      let (same,different) = List.partition split_fun ids in
      (* proceed by also splitting the ids that were different from 
	 node0 *)
      same :: (split_colors diag col different)

(* Given a diagram and a coloring, return a new coloring which has more
   colors and distinguishes more ids with non-equivalent nodes. *)
let rec refine_coloring diag col =
  List.flatten (List.map (split_colors diag col) col)   

(* starts with the coloring that gives all nodes the same color, and
   then successively refines it until the number of colors no longer
   increases. It returns the coloring with the least possible number of
   colors. *)
 
let minimal_coloring diag =
  let rec aux diag col n = (* n is the number of colors in col *)
    let col' = refine_coloring diag col in
    let n' = List.length col' in (* number of colors in col' *)
    if (n=n') then col else
    aux diag col' n'
  in
  let ids = Library.get_domain diag in
  aux diag [ids] 1

(* identify equivalent colors in a diagram, replacing each color by the first
   equivalent color mentioned in the coloring. *)

let rec identify_colors col diag =
  (* replace any occurences of an id in slackids in node with origid *)
  let replace_in_node origid slackids node =
    let f = fun ptr ->
      match ptr with
        Var(tvar) -> ptr
      |	Id(id) -> if (List.mem id slackids) then (Id(origid))
                  else ptr
    in
    match node with
      Int -> Int
    |  Bool -> Bool
    |  Arrow(ptr1,ptr2) -> Arrow(f ptr1, f ptr2)
    |  Prod(ptrlist) -> Prod(List.map f ptrlist)
    |  Sum(ptrlist) -> Sum(List.map f ptrlist)
  in
 
(* replace each occurence of the ids in slackids with origid in the
   diagram *)
  let replace_references origid slackids diag =
    let (idlist,nodelist) = List.split diag in
    let nodelist' = List.map (replace_in_node origid slackids) nodelist in
    List.combine idlist nodelist'
  in
  (* remove all the (id,node) entries from the diagram where each id is in
     ids *)
  let rec remove_entries diag ids =
    match ids with
      [] -> diag
    | hd::tl ->
        let diag' = List.remove_assoc hd diag in
        remove_entries diag' tl
  in
  match col with
    [] -> diag
  | []::tl -> identify_colors tl diag (* this should never happen *)
  | (first::slack)::tl ->
      let diag' = remove_entries diag slack in
      let diag'' = replace_references first slack diag' in
      identify_colors tl diag''

(* return a diagram equivalent to the one given, obtained by identifying
   nodes of the same color and replacing each color by the first equilvanet
   color mentioned in the coloring.  here use the minimal coloring. *)
let normalize_diagram diag =
  let col = minimal_coloring diag in
  identify_colors col diag

(* identify the color that an id is a member of and return the first element
   in that list.  if the id is not in any color, return that id. *)
let rec find id col =
  match col with
    [] -> Id(id) (* never happens *)
  | hd::tl ->
      if (List.mem id hd) then (Id(List.hd hd))
      else find id tl
	  
(* given a typeexp (ptr, diag).  if ptr is a tvar then return it together
   with the empty diagram.  else (1) find the minimal coloring for diag and
   normalize it and (2) if ptr is not present in the normalized diagram, 
   return a ptr which is isomorphic to it. *)

let normalize_typeexp typeexp =
  let (ptr,diag) = typeexp in
  match ptr with
    Var(tvar) -> (ptr,[]) (* throw away non-rooted diagram *)
  | Id(id) ->
      let col = minimal_coloring diag in
      let ptr' = find id col in (* for t3 this is "Id 1" *)
      let diag' = identify_colors col diag in
      (ptr',diag')
	
(* test vars *)
let id1=Id(1);; 
let id2=Id(2);; 
let id3=Id(3);;
let var1=Var(Usertvar("var1"));; 
let var2=Var(Usertvar("var2"));;   
let var3=Var(Usertvar("var3"));; 
let diag1=[(1,Arrow(var1,id1))];;  
let diag2=[(1,Arrow(var1,id2));(2,Arrow(var2,id1))];; 
let diag3=[(1,Arrow(var1,id2));(2,Bool);(3,Arrow(var1,id2))];;
let t1 = (Id 1, [(1, Prod [Id 2; Id 1; Id 3]); (2, Arrow (Id 1, Var 
(Usertvar "alpha"))); (3, Bool)]);;
let t2 = (Id 2, [(1, Prod [Id 2; Id 1; Id 3]); (2, Arrow (Id 1, Var 
(Usertvar "alpha"))); (3, Bool)]);;
let t3=(Id 3, diag3);;
let t4=(Id 1, diag3);;
